;; Generated package description from org-mem.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "org-mem" "0.33.0.0.20260303.75813" "Fast info from a large number of Org file contents" '((emacs "29.1") (el-job "2.7.3") (llama "1.0") (truename-cache "0.3.2")) :commit "3a90968f693dfa80d6c8b2b3f3482de63a390f7d" :authors '(("Martin Edström" . "meedstrom@runbox.eu")) :maintainer '("Martin Edström" . "meedstrom@runbox.eu") :keywords '("text") :url "https://github.com/meedstrom/org-mem")
